#!/usr/bin/python
# encoding: utf-8
import time
import os
import requests
from ftplib import FTP
from ftplib import FTP_TLS
import hashlib
import ssl
import sys
import subprocess
import json


'''
# 1. 计算并获取md5值
# 2. 创建软连接
# 3. 关闭虚拟机
# 4. 获取原始磁盘的路径
 # 5. 将磁盘文件重命名为 .old
 # 6. 下载模板文件
 # 7. 检查模板文件的 MD5 值
 # 8. 比较模板文件的 MD5 值
 # 9. 挂载cdrom
 # 10. 启动虚拟机
 # 11. 手动将 cdrom 踢掉
 # 12. 手动 删除 old 文件
'''

def caculate_template_md5(template_filename):
    url = "https://download-md5.cloudclusters.io/submit_md5"
    headers = {
        "Content-Type": "application/json"
    }
    auth = ("LIMIT", "LIMIT")  # 用户名和密码
    data = {
        "file_name": template_filename
    }

    try:
        # 发送 POST 请求
        response = requests.post(url, headers=headers, auth=auth, json=data)

        # 检查响应状态码
        if response.status_code in [200,201,202]:
            return response.json()  # 返回响应的 JSON 数据
        else:
            print(f"request failed: {response.status_code}")
            print(f"body: {response.text}")
            return None
    except Exception as e:
        print(f"request error: {e}")
        return None


def get_template_md5(template_filename, max_retries=20, retry_interval=30):
    """
    不断重试获取文件的 MD5 值，直到请求成功（状态码为 200 或 201），或达到最大重试次数。

    :param template_filename: 文件名
    :param max_retries: 最大重试次数，默认 10 次
    :param retry_interval: 每次重试的时间间隔（秒），默认 30 秒
    :return: 返回 MD5 值的 JSON 数据，或者 None（如果最终失败）
    """
    url = "https://download-md5.cloudclusters.io/check_md5"
    headers = {
        "Content-Type": "application/json"
    }
    auth = ("LIMIT", "LIMIT")  # 用户名和密码
    data = {
        "file_name": template_filename
    }

    for attempt in range(1, max_retries + 1):
        try:
            # 发送 GET 请求
            response = requests.post(url, headers=headers, auth=auth, json=data)

            # 检查响应状态码
            if response.status_code in [200, 201]:
                print(f"get MD5 value: {response.json()}")
                return response.json()  # 返回响应的 JSON 数据
            else:
                print(f"request failed，status code: {response.status_code}, retry {attempt}/{max_retries}")
                print(f"body: {response.text}")
        except Exception as e:
            print(f"request error: {e}, retry {attempt}/{max_retries}")

        # 如果未达到最大重试次数，等待一段时间后重试
        if attempt < max_retries:
            print(f"wait {retry_interval} sceond retry...")
            time.sleep(retry_interval)

    print("get MD5 failed")
    return None


def create_symlink(file_name):
    """
    调用远程 API 创建软链接。

    :param file_name: 文件名（用于创建软链接的源文件）
    :return: API 返回的 JSON 数据或 None（如果失败）
    """
    url = "https://download-md5.cloudclusters.io/create_symlink"
    headers = {
        "Content-Type": "application/json"
    }
    auth = ("LIMIT", "LIMIT")  # 用户名和密码
    data = {
        "file_name": file_name
    }

    try:
        # 发送 POST 请求
        response = requests.post(url, headers=headers, auth=auth, json=data)

        # 检查响应状态码
        if response.status_code in [200, 201]:
            print("soft link create success！")
            return response.json()  # 返回响应的 JSON 数据
        else:
            print(f"status_code: {response.status_code}")
            print(f"body: {response.text}")
            return None

    except Exception as e:
        print(f"error: {e}")
        return None


def shutdown_hyperv_vm(vm_name):
    """
    关闭指定的 Hyper-V 虚拟机。

    :param vm_name: 虚拟机的名称
    :return: 返回关闭操作的结果信息
    """
    try:
        # PowerShell 命令关闭虚拟机
        command = f"Stop-VM -Name \"{vm_name}\" -Force"

        # 使用 subprocess 调用 PowerShell
        result = subprocess.run(
            ["powershell", "-Command", command],
            capture_output=True,
            text=True
        )

        # 检查命令执行结果
        if result.returncode == 0:
            return f"vm {vm_name} stop success"
        else:
            return f"shutdown {vm_name} error：{result.stderr.strip()}"
    except Exception as e:
        return f"error：{str(e)}"



def get_scsi_disk_path(vm_name, controller_number=0, controller_location=0):
    """
    获取指定 Hyper-V 虚拟机的 SCSI 控制器上的磁盘路径。

    :param vm_name: 虚拟机的名称
    :param controller_number: SCSI 控制器编号（默认为 0）
    :param controller_location: SCSI 控制器位置（默认为 0）
    :return: 磁盘路径或错误信息
    """
    try:
        # PowerShell 脚本过滤 SCSI 0.0 的磁盘
        ps_script = f"""
        $vm = Get-VM -Name "{vm_name}"
        if ($null -eq $vm) {{
            Write-Output @{{Error = 'VM not found'}}
            exit 1
        }}
        $drive = Get-VMHardDiskDrive -VMName "{vm_name}" | Where-Object {{
            $_.ControllerType -eq 'SCSI' -and
            $_.ControllerNumber -eq {controller_number} -and
            $_.ControllerLocation -eq {controller_location}
        }}
        if ($null -eq $drive) {{
            Write-Output @{{Error = 'Disk not found'}}
            exit 1
        }}
        $vhd = Get-VHD -Path $drive.Path
        [PSCustomObject]@{{
            Path = $vhd.Path
            Size = $vhd.Size
            ControllerType = $drive.ControllerType
            ControllerNumber = $drive.ControllerNumber
            ControllerLocation = $drive.ControllerLocation
        }} | ConvertTo-Json -Depth 3
        """

        # 使用 subprocess 调用 PowerShell
        result = subprocess.run(
            ["powershell", "-Command", ps_script],
            capture_output=True,
            text=True
        )

        if result.returncode != 0:
            return f"get disk path error：{result.stderr.strip()}"

        # 解析 JSON 输出
        output = result.stdout.strip()
        try:
            disk_info = json.loads(output)
            return disk_info  # 返回磁盘详细信息
        except json.JSONDecodeError:
            return f"get disk info error：{output}"

    except Exception as e:
        return f"error：{str(e)}"


def rename_disk_to_old(disk_path):
    """
    将指定的磁盘文件重命名为 .old 后缀。

    :param disk_path: 原始磁盘文件路径
    :return: 新的文件路径
    """
    if not os.path.exists(disk_path):
        raise FileNotFoundError(f"disk file not found: {disk_path}")

    # 构造新的文件路径
    new_path = f"{disk_path}.old"

    # 检查目标文件是否已经存在
    if os.path.exists(new_path):
        print(f"file exists: {new_path}")
        return new_path
    # 重命名文件
    os.rename(disk_path, new_path)
    print(f"file rename success: {disk_path} -> {new_path}")
    return new_path


class MyFTP(FTP):
    '''
    connect to FTP Server
    '''
    def __init__(self):
        print('make a object')


    def ConnectFTP(self, remote_ip="", remote_port=21, login_name="", login_password=""):
        # ftp = FTP()
        ftp = FTP_TLS()
        ftp.ssl_version = ssl.PROTOCOL_TLSv1_2

        try:
            ftp.connect(remote_ip, remote_port, 600)
            print('Connect Success')
        except Exception as e:
            print(sys.stderr, "Coonnect Failed - %s" %e)
            return (0, 'Connect Failed')
        else:
            try:
                ftp.login(login_name, login_password)
                print('login Success')
            except Exception as e:
                print(sys.stderr, "Login Failed - %s" %e)
                return (0, 'Login Faild')
            else:
                print('retrun 1')
                return (1, ftp)


    def Download(self, remoteHost="", remotePort="", login_name="", login_password="", remotePath="", localPath=""):
        # connect to the FTP Server and check the return
        res = self.ConnectFTP(remote_ip=remoteHost, remote_port=remotePort, login_name=login_name, login_password=login_password)
        if (res[0] != 1):
            print(sys.stderr, res[1])
            sys.exit(1)


        # change the remote directory and get the remote file size
        ftp = res[1]
        ftp.set_pasv(1)
        dires = self.splitpath(remotePath)
        # if dires[0]:
        #     # change remote work dir
        #     ftp.cwd(dires[0])
        remotefile = dires[1]
        print(dires[0],'---',dires[1])
        fsize = ftp.size(remotefile)
        if fsize == 0:
            return

        # check if local file exists and get the local file size
        lsize = 0
        if os.path.exists(localPath):
            lsize = os.stat(localPath).st_size

        if lsize >= fsize:
            print('local file is bigger or equal remote file')
            return

        blocksize = 1024*1024
        cmpsize = lsize
        ftp.voidcmd('TYPE I')
        conn = ftp.transfercmd('RETR ' + remotefile, lsize)
        lwrite = open(localPath, 'ab')
        while True:
            data = conn.recv(blocksize)
            if not data:
                break
            lwrite.write(data)
            cmpsize += len(data)
            print('\b'*30,'download process:%.2f%%'%(float(cmpsize)/fsize*100))
        lwrite.close()
        ftp.voidcmd('NOOP')
        ftp.voidresp()
        conn.close()


    def splitpath(self, remotepath):
        position = remotepath.rfind('/')
        return (remotepath[:position + 1], remotepath[position + 1:])

    def md5_calc(self, file):
        md5_value = hashlib.md5()
        with open(file, 'rb') as file_b:
            while True:
                data_flow = file_b.read(8096)  # 每次读入8089kb进入内存
                if not data_flow:  # 读取完后返回空值，False
                    break
                md5_value.update(data_flow)
        file_b.close()
        return md5_value.hexdigest()

    def get_filesize(self, filename):
        fsize = os.path.getsize(filename)
        return fsize


def download_file_from_ftp(login_name, login_password, remotefile, localPath):
    try:
        ftpobj = MyFTP()
        ftpobj.Download(
            remoteHost='165.73.249.37',
            remotePort=10021,
            login_name=login_name,
            login_password=login_password,
            remotePath="/ftp-download/" + remotefile,
            localPath=localPath
        )
    except Exception as e:
        print(f"download file failed: {e}")
        return False


def calculate_md5(file_path):
    """
    计算文件的 MD5 值。
    :param file_path: 文件的完整路径。
    :return: 文件的 MD5 哈希值。
    """
    md5_hash = hashlib.md5()
    with open(file_path, "rb") as file:
        chunk = file.read(8192)
        while chunk:
            md5_hash.update(chunk)
            chunk = file.read(8192)
    return md5_hash.hexdigest()


def find_cdrom_path(disk_path):
    """
    根据磁盘路径推导出 CD-ROM 的 ISO 路径。

    :param disk_path: 虚拟机磁盘路径
    :return: ISO 文件路径
    """
    # 获取磁盘目录
    base_dir = os.path.dirname(os.path.dirname(disk_path))
    cdrom_path = os.path.join(base_dir, "config_drive", "config_drive.iso")

    if not os.path.exists(cdrom_path):
        raise FileNotFoundError(f"ISO not found: {cdrom_path}")

    return cdrom_path


def attach_cdrom(vm_name, iso_path):
    """
    挂载 ISO 文件到虚拟机的 CD-ROM。

    :param vm_name: 虚拟机名称
    :param iso_path: ISO 文件路径
    """
    # 检查 ISO 文件是否存在
    if not os.path.exists(iso_path):
        raise FileNotFoundError(f"ISO not found: {iso_path}")

    # PowerShell 脚本
    powershell_script = f"""
    $VMName = "{vm_name}"
    $GenerateConfigDriveISOFile = "{iso_path}"

    $DVDDrive = Get-VMDvdDrive -VMName $VMName

    if (!$DVDDrive) {{
        Add-VMDvdDrive -VMName $VMName
    }}

    $DVDDrive = Get-VMDvdDrive -VMName $VMName
    Set-VMDvdDrive -VMName $VMName -Path $GenerateConfigDriveISOFile -ControllerNumber $DVDDrive.ControllerNumber -ControllerLocation $DVDDrive.ControllerLocation

    Write-Output "Attach CDROM success."
    """

    try:
        # 运行 PowerShell 脚本
        result = subprocess.run(
            ["powershell", "-Command", powershell_script],
            check=True,
            text=True,
            capture_output=True,
        )
        print(result.stdout)
        print(f"ISO file attach success {vm_name}: {iso_path}")
    except subprocess.CalledProcessError as e:
        print(f"mount ISO file failed: {e.stderr}")
        raise


def start_virtual_machine(vm_name):
    """
    启动指定的 Hyper-V 虚拟机。

    :param vm_name: 虚拟机名称
    """
    powershell_script = f"""
    $VMName = "{vm_name}"

    # 检查虚拟机是否存在
    $VM = Get-VM -Name $VMName -ErrorAction SilentlyContinue
    if (-not $VM) {{
        Write-Output "虚拟机 $VMName 不存在。"
        exit 1
    }}

    # 检查虚拟机状态
    if ($VM.State -eq "Running") {{
        Write-Output "虚拟机 $VMName 已经在运行中。"
    }} else {{
        Start-VM -Name $VMName
        Write-Output "machine $VMName start success."
    }}
    """

    try:
        # 执行 PowerShell 脚本
        result = subprocess.run(
            ["powershell", "-Command", powershell_script],
            check=True,
            text=True,
            capture_output=True,
        )
        print(result.stdout)
    except subprocess.CalledProcessError as e:
        print(f"start vm error: {e.stderr}")
        raise


if __name__ == "__main__":
    if os.name == 'nt':
        os.system('')

    vm_name = input("Please input the vm name: [example: vps_123]:").strip()
    template_file = input("Please input the template name: [example vps_123-win2016.vhdx]: ").strip()
    ftp_user = input("Please input FTP username: [example xxx]: ").strip()
    ftp_password = input("Please input FTP password: [example xxx]: ").strip()

    process_len = "12"
    # 0. 连接ftp
    ftp_obj = MyFTP()
    # 尝试连接到 FTP 服务器
    status, ftp = ftp_obj.ConnectFTP(
        remote_ip="165.73.249.37",
        remote_port=10021,
        login_name=ftp_user,
        login_password=ftp_password
    )
    if status == 1:
        print("Successfully connected and logged into the FTP server.")
        ftp.quit()  # 结束会话
    else:
        print(f"Failed to connect or login: {ftp}")
        input("Press Enter to exit.")
        sys.exit(1)

    # 1. 获取md5值
    try:
        print(f"\033[32m1. get file MD5 value, [1/{process_len}]\033[0m")
        caculate_template_md5(template_file)
        # 获取模板文件的 MD5 值
        templateinfo = get_template_md5(template_file)
        template_md5 = templateinfo["md5"]
        print(f"\033[32mtemplate {template_file} MD5 is：{template_md5}\033[0m")
    except:
        print(f"\033[31mget template {template_file} MD5 failed\033[0m")

    # 2. 创建软连接
    print(f"\033[32m2. create {template_file} soft link, [2/{process_len}]\033[0m")
    response_data = create_symlink(template_file)
    softlinkfile = "ln-vps-tmpl-" + template_file
    print(f"\033[32mlink create success，the name is：{softlinkfile}\033[0m")

    # 3. 关闭虚拟机
    print(f"\033[32m3. shutdown vm {vm_name}, [3/{process_len}]\033[0m")
    shutdown_hyperv_vm(vm_name=vm_name)

    # 4. 获取原始磁盘的路径
    print(f"\033[32m4. get {vm_name} raw disk path, [4/{process_len}]\033[0m")
    raw_disk_info = get_scsi_disk_path(vm_name=vm_name)
    raw_disk_path = raw_disk_info["Path"]
    print(f"\033[32mvm {vm_name} raw disk is：{raw_disk_path}\033[0m")

    # 5. 将磁盘文件重命名为 .old
    print(f"\033[32m5. rename {vm_name} file to .old, [5/{process_len}]\033[0m")
    oldbackup_file_path = rename_disk_to_old(disk_path=raw_disk_path)
    print(f"\033[32mvm {vm_name} raw disk {raw_disk_path} rename to：{oldbackup_file_path}\033[0m")

    time.sleep(10)
    # 6. 下载模板文件
    print(f"\033[32m6. download {template_file}, [6/{process_len}]\033[0m")
    download_file_from_ftp(login_name=ftp_user, login_password=ftp_password, remotefile=softlinkfile,
                           localPath=raw_disk_path)
    print(f"\033[32mtemplate file {template_file} download finished，saved as：{raw_disk_path}\033[0m")


    # 7. 检查模板文件的 MD5 值
    print(f"\033[32m7. check md5 sum {raw_disk_path}, [7/{process_len}]\033[0m")
    newfile_md5_value = calculate_md5(raw_disk_path)
    print(f"\033[32mthe file {raw_disk_path} md5 is：{newfile_md5_value}\033[0m")

    # 8. 比较模板文件的 MD5 值
    print(f"\033[32m8. compare MD5 value, [8/{process_len}]\033[0m")
    if template_md5 == newfile_md5_value:
        print("\033[32m MD5 compare success\033[0m")
    else:
        print("\033[31m MD5 compare failed\033[0m")
        print("\033[31mexit\033[0m")
        input("Upload complete. Press Enter to exit...")
        sys.exit(1)

    # 9. 挂载 Cdrom
    print(f"\033[32m9. mount Cdrom, [9/{process_len}]\033[0m")
    cdrom_file = find_cdrom_path(disk_path=raw_disk_path)
    print(f"\033[32mvm {vm_name} CDROM path is：{cdrom_file}\033[0m")
    attach_cdrom(vm_name=vm_name, iso_path=cdrom_file)
    print(f"\033[32mvm {vm_name}  CDROM mount success！\033[0m")

    # 10. 启动虚拟机
    print(f"\033[32m10. start vm  {vm_name}, [10/{process_len}]\033[0m")
    start_virtual_machine(vm_name=vm_name)

    print(f"\033[33mPlease manually check the running status of the virtual machine {vm_name}, including IP, hostname, password, etc.\033[0m")

    # 11. 手动踢掉 cdrom
    print(f"\033[31m11. please manually kick off the cdrom file {cdrom_file} of {vm_name}, [11/{process_len}]\033[0m")

    # 12. 手动删除 old 文件
    print(f"\033[31m12. please manually delete the old file {oldbackup_file_path} of {vm_name}, [12/{process_len}]\033[0m")

    input("Upload complete. Press Enter to exit...")
    input("Upload complete. Press Enter to exit...")
